async def empty_asgi_handler(scope, receive, send):
    if scope["type"] == "lifespan":
        while True:
            message = await receive()
            if message["type"] == "lifespan.startup":
                await send({"type": "lifespan.startup.complete"})
                return
            elif message["type"] == "lifespan.shutdown":
                await send({"type": "lifespan.shutdown.complete"})
                return

    await send(
        {
            "type": "http.response.start",
            "status": 404,
            "headers": [(b"content-length", b"0")],
        }
    )
    await send({"type": "http.response.body"})



import asyncio
import functools
import typing

from . import asgitypes

MAX_QUEUE_SIZE = 10

T = typing.TypeVar("T")
U = typing.TypeVar("U")


class DispatcherMiddleware:
    def __init__(self, mounts: dict[str, asgitypes.ASGI3Application]) -> None:
        self.mounts = mounts

    async def __call__(self, scope: asgitypes.Scope, receive: typing.Callable, send: typing.Callable) -> None:
        if scope["type"] == "lifespan":
            await self._handle_lifespan(scope, receive, send)
        else:
            for path, app in self.mounts.items():
                if scope["path"].startswith(path):
                    scope["path"] = scope["path"][len(path) :] or "/"
                    return await app(scope, receive, send)

            if scope["type"] == "http":
                await send(
                    {
                        "type": "http.response.start",
                        "status": 404,
                        "headers": [(b"content-length", b"0")],
                    }
                )
                await send({"type": "http.response.body"})
            elif scope["type"] == "websocket":
                await send({"type": "websocket.close"})

    async def _handle_lifespan(
        self,
        scope: asgitypes.Scope,
        receive: asgitypes.ASGIReceiveCallable,
        send: asgitypes.ASGISendCallable,
    ) -> None:
        self.app_queues = {path: asyncio.Queue(MAX_QUEUE_SIZE) for path in self.mounts}
        self.startup_complete = dict.fromkeys(self.mounts, False)
        self.shutdown_complete = dict.fromkeys(self.mounts, False)

        tasks = []
        try:
            for path, app in self.mounts.items():
                tasks.append(
                    asyncio.create_task(
                        app(
                            scope,
                            self.app_queues[path].get,
                            functools.partial(self.send, path, send),  # type: ignore
                        )
                    )
                )

            while True:
                message = await receive()
                for queue in self.app_queues.values():
                    await queue.put(message)
                if message["type"] == "lifespan.shutdown":
                    break
        finally:
            await asyncio.wait(tasks)

    async def send(self, path: str, send: typing.Callable, message: dict) -> None:
        if message["type"] == "lifespan.startup.complete":
            self.startup_complete[path] = True
            if all(self.startup_complete.values()):
                await send({"type": "lifespan.startup.complete"})
        elif message["type"] == "lifespan.shutdown.complete":
            self.shutdown_complete[path] = True
            if all(self.shutdown_complete.values()):
                await send({"type": "lifespan.shutdown.complete"})



import logging

from loguru import logger

try:
    from sqlalchemy.log import Identified, _qual_logger_name_for_cls
except ImportError:
    raise ImportError(
        "dependency 'sqlalchemy' is required for sqlalchemy service\nplease install it or install 'graia-amnesia[sqla]'"
    )

from ..utils import LoguruHandler, get_subclasses
from .model import Base as Base
from .service import SqlalchemyService as SqlalchemyService


def patch_logger(log_level: str | int = "INFO", sqlalchemy_echo: bool = False) -> None:
    handler = LoguruHandler()
    logging.getLogger("sqlalchemy").addHandler(handler)

    if isinstance(log_level, str):
        log_level = logger.level(log_level).no

    echo_log_level = log_level if sqlalchemy_echo else logging.WARNING

    levels = {
        "alembic": log_level,
        "sqlalchemy": log_level,
        **{_qual_logger_name_for_cls(cls): echo_log_level for cls in set(get_subclasses(Identified))},
    }

    for name, level in levels.items():
        logging.getLogger(name).setLevel(level)



import asyncio
from datetime import timedelta
from heapq import heappop, heappush
from time import time
from typing import Any

from launart import Launart, Service
from launart.status import Phase


class Memcache:
    cache: dict[str, tuple[float | None, Any]]
    expire: list[tuple[float, str]]

    def __init__(
        self,
        cache: dict[str, tuple[float | None, Any]],
        expire: list[tuple[float, str]],
    ):
        self.cache = cache
        self.expire = expire

    async def get(self, key: str, default: Any = None) -> Any:
        value = self.cache.get(key)
        if value:
            if value[0] is None or value[0] >= time():
                return value[1]
            del self.cache[key]
        return default

    async def set(self, key: str, value: Any, expire: timedelta | None = None) -> None:
        if expire is None:
            self.cache[key] = (None, value)
        else:
            expire_time = time() + expire.total_seconds()
            self.cache[key] = (expire_time, value)
            heappush(self.expire, (expire_time, key))

    async def delete(self, key: str, strict: bool = False) -> None:
        if strict or key in self.cache:
            del self.cache[key]

    async def clear(self) -> None:
        self.cache.clear()
        self.expire.clear()

    async def has(self, key: str) -> bool:
        return key in self.cache

    async def keys(self) -> list[str]:
        return list(self.cache.keys())


class MemcacheService(Service):
    id = "cache.client/memcache"

    interval: float
    _cache: dict[str, tuple[float | None, Any]]
    expire: list[tuple[float, str]]

    def __init__(
        self,
        interval: float = 0.1,
        cache: dict[str, tuple[float | None, Any]] | None = None,
        expire: list[tuple[float, str]] | None = None,
    ):
        self.interval = interval
        self._cache = cache or {}
        self.expire = expire or []
        super().__init__()

    @property
    def required(self):
        return set()

    @property
    def stages(self) -> set[Phase]:
        return {"blocking"}

    @property
    def cache(self):
        return Memcache(self._cache, self.expire)

    async def launch(self, manager: Launart) -> None:
        async with self.stage("blocking"):
            while not manager.status.exiting:
                while self.expire and self.expire[0][0] <= time():
                    _, key = heappop(self.expire)
                    self._cache.pop(key, None)
                await asyncio.sleep(self.interval)



try:
    from .uvicorn import UvicornASGIService as _UvicornASGIService
except ImportError:
    _UvicornASGIService = None

try:
    from .hypercorn import HypercornASGIService as _HypercornASGIService
except ImportError:
    _HypercornASGIService = None


def __getattr__(name):
    if name == "UvicornASGIService":
        if _UvicornASGIService is None:
            raise ImportError("Please install `uvicorn` first. Install with `pip install graia-amnesia[uvicorn]`")
        return _UvicornASGIService
    if name == "HypercornASGIService":
        if _HypercornASGIService is None:
            raise ImportError("Please install `hypercorn` first. Install with `pip install graia-amnesia[hypercorn]`")
        return _HypercornASGIService
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



from collections.abc import Callable
from typing import Any, Literal
from typing_extensions import TypedDict

from sqlalchemy.engine.interfaces import IsolationLevel, _ExecuteOptions, _ParamStyle
from sqlalchemy.log import _EchoFlagType
from sqlalchemy.pool import Pool, _CreatorFnType, _CreatorWRecFnType, _ResetStyleArgType


class EngineOptions(TypedDict, total=False):
    connect_args: dict[Any, Any]
    convert_unicode: bool
    creator: _CreatorFnType | _CreatorWRecFnType
    echo: _EchoFlagType
    echo_pool: _EchoFlagType
    enable_from_linting: bool
    execution_options: _ExecuteOptions
    future: Literal[True]
    hide_parameters: bool
    implicit_returning: Literal[True]
    insertmanyvalues_page_size: int
    isolation_level: IsolationLevel
    json_deserializer: Callable[..., Any]
    json_serializer: Callable[..., Any]
    label_length: int | None
    logging_name: str
    max_identifier_length: int | None
    max_overflow: int
    module: Any | None
    paramstyle: _ParamStyle | None
    pool: Pool | None
    poolclass: type[Pool] | None
    pool_logging_name: str
    pool_pre_ping: bool
    pool_size: int
    pool_recycle: int
    pool_reset_on_return: _ResetStyleArgType | None
    pool_timeout: float
    pool_use_lifo: bool
    plugins: list[str]
    query_cache_size: int
    use_insertmanyvalues: bool
    kwargs: dict[str, Any]



from typing import cast

from launart import Launart, Service
from launart.status import Phase

try:
    from httpx import AsyncClient, Timeout
except ImportError:
    raise ImportError(
        "dependency 'httpx' is required for httpx client service\nplease install it or install 'graia-amnesia[httpx]'"
    )


class HttpxClientService(Service):
    id = "http.client/httpx"
    session: AsyncClient

    def __init__(self, session: AsyncClient | None = None) -> None:
        self.session = cast(AsyncClient, session)
        super().__init__()

    @property
    def stages(self) -> set[Phase]:
        return {"preparing", "cleanup"}

    @property
    def required(self):
        return set()

    async def launch(self, manager: Launart):
        async with self.stage("preparing"):
            if self.session is None:
                self.session = AsyncClient(timeout=Timeout())
        async with self.stage("cleanup"):
            await self.session.aclose()



from __future__ import annotations

from collections.abc import Iterable, Iterator, Sequence
from copy import deepcopy
from typing import TYPE_CHECKING, Any, TypeVar, overload
from typing_extensions import Self

from .element import Element, Text

if TYPE_CHECKING:
    E = TypeVar("E", bound=Element)


class MessageChain:
    """即 "消息链", 被用于承载整个消息内容的数据结构, 包含有一有序列表, 包含有继承了 Element 的各式类实例.

    - 你可以通过实例化 `MessageChain` 创建一个消息链

    - `str(chain)`: 获取字符串形式表示的消息

    - `has`: 判断特定的内容是否存在于消息链中

    - `get`: 获取消息链中的所有特定类型的元素

    - `get_first`: 获取消息链中的第 1 个特定类型的元素

    - `get_one`: 获取消息链中的第 index + 1 个特定类型的元素

    - `merge`: 将消息链中相邻的 Text 元素合并为一个 Text 元素.

    - `startswith`: 判断消息链是否以指定的文本开头

    - `endswith`: 判断消息链是否以指定的文本结尾

    - `include`: 创建只包含指定元素类型的消息链

    - `exclude`: 创建排除指定元素类型的消息链

    - `split`: 用指定文本将消息链拆分为多个

    - `append`: 将指定的元素添加到消息链的末尾

    - `extend`: 将指定的序列/消息链添加到消息链的末尾

    - `only`: 检查是否只包含指定元素

    - `index`: 获取指定元素在消息链中的索引

    - `index_sub`: 搜索子消息链在本消息链的出现位置

    - `count`: 获取消息链中指定元素的数量

    - `copy`: 获取消息链的拷贝

    - `chain.join(chains)`: 拼接多个消息链并插入指定内容

    - `strip`: 删除前导和尾随空白字符。如果给出了 elements 而不是 None，则依据 elements 中的字符或元素进行删除。

    - `lstrip`: 删除前导空白字符。如果给出了 elements 而不是 None，则依据 elements 中的字符或元素进行删除。

    - `rstrip`: 删除尾随空白字符。如果给出了 elements 而不是 None，则依据 elements 中的字符或元素进行删除。

    - `removeprefix`: 尝试移除前缀

    - `removesuffix`: 尝试移除后缀

    - `replace`: 替换消息链的一部分

    """

    _text_class: type[Text]

    content: list[Element]

    def __init__(self, elements: Sequence[str | Element]):
        """从传入的序列(可以是元组 tuple, 也可以是列表 list) 创建消息链.
        Args:
            elements (Sequence[str | Element]): 包含且仅包含消息元素和字符串的序列
        Returns:
            MessageChain: 以传入的序列作为所承载消息的消息链
        """
        self.content = []
        if isinstance(elements, str):
            self.content.append(self._text_class(elements))
        else:
            for element in elements:
                if isinstance(element, str):
                    element = self._text_class(element)
                self.content.append(element)

    def has(self, item: Element | type[Element] | Self | Sequence[str | Element]) -> bool:
        """
        判断消息链中是否含有特定的内容.

        Args:
            item (Element | type[Element] | Self | Sequence[str | Element]): 需判断内容.

        Returns:
            bool: 判断结果
        """
        if isinstance(item, type):
            return item in [type(i) for i in self.content]
        if isinstance(item, Element):
            return item in self.merge().content
        if isinstance(item, (Sequence, MessageChain)):
            return bool(self.index_sub(item))

        raise ValueError(f"{item} is not an acceptable argument!")

    def get(self, element_class: type[E], count: int = -1) -> list[E]:
        """
        获取消息链中所有特定类型的消息元素

        Args:
            element_class (type[E]): 指定的消息元素的类型, 例如 "Plain", "At", "Image" 等.
            count (int, optional): 至多获取的元素个数

        Returns:
            list[E]: 获取到的符合要求的所有消息元素; 另: 可能是空列表([]).
        """
        if count == -1:
            count = len(self.content)
        return [i for i in self.content if isinstance(i, element_class)][:count]

    def get_one(self, element_class: type[E], index: int) -> E:
        """获取消息链中第 index + 1 个特定类型的消息元素
        Args:
            element_class (type[Element]): 指定的消息元素的类型, 例如 "Text", "Notice", "Image" 等.
            index (int): 索引, 从 0 开始数
        Returns:
            T: 消息链第 index + 1 个特定类型的消息元素
        """
        return self.get(element_class)[index]

    def get_first(self, element_class: type[E]) -> E:
        """获取消息链中第 1 个特定类型的消息元素
        Args:
            element_class (type[Element]): 指定的消息元素的类型, 例如 "Text", "Notice", "Image" 等.
        Returns:
            T: 消息链第 1 个特定类型的消息元素
        """
        return self.get(element_class)[0]

    def __str__(self) -> str:
        """获取以字符串形式表示的消息链, 且趋于通常你见到的样子.
        Returns:
            str: 以字符串形式表示的消息链
        """
        return "".join(str(i) for i in self.content)

    def join(self, *chains: Self | Iterable[Self]) -> Self:
        """将多个消息链连接起来, 并在其中插入自身.

        Args:
            *chains (Iterable[MessageChain]): 要连接的消息链.

        Returns:
            MessageChain: 连接后的消息链, 已对文本进行合并.
        """
        result: list[Element] = []
        list_chains: list[MessageChain] = []
        for chain in chains:
            if isinstance(chain, MessageChain):
                list_chains.append(chain)
            else:
                list_chains.extend(chain)

        for chain in list_chains:
            if chain is not list_chains[0]:
                result.extend(deepcopy(self.content))
            result.extend(deepcopy(chain.content))
        return self.__class__(result).merge()

    __contains__ = has

    @overload
    def __getitem__(self, item: type[E]) -> list[E]: ...

    @overload
    def __getitem__(self, item: int) -> Element: ...

    @overload
    def __getitem__(self, item: slice) -> Self: ...

    def __getitem__(self, item: type[Element] | int | slice) -> Any:
        """取出子消息链, 或元素.

        通过 `type` 取出属于 `type` 的元素列表

        通过 `int` 取出对应位置元素.

        通过 `slice` 取出子消息链.

        Args:
            item (type[Element] | int | slice): 索引项

        Returns:
            list[Element] | Element | MessageChain: 索引结果.
        """
        if isinstance(item, slice):
            return self.__class__(self.content[item])
        if isinstance(item, int):
            return self.content[item]
        elif issubclass(item, Element):
            return self.get(item)
        else:
            raise NotImplementedError(f"{type(item)} is not allowed for item getting")

    def merge(self, *, copy: bool = True) -> Self:
        """合并相邻的 Text 项, 选择返回一个新的消息链实例

        Returns:
            MessageChain: 得到的新的消息链实例, 里面不应存在有任何的相邻的 Text 元素.
        """

        result = []

        texts = []
        for i in self.content:
            if not isinstance(i, Text):
                if texts:
                    result.append(self._text_class("".join(texts)))
                    texts.clear()  # 清空缓存
                result.append(i)
            else:
                texts.append(i.text)
        if texts:
            result.append(self._text_class("".join(texts)))
            texts.clear()  # 清空缓存
        if copy:
            return self.__class__(result)
        self.content.clear()
        self.content.extend(result)
        return self

    def exclude(self, *types: type[Element]) -> Self:
        """将除了在给出的消息元素类型中符合的消息元素重新包装为一个新的消息链
        Args:
            *types (type[Element]): 将排除在外的消息元素类型
        Returns:
            MessageChain: 返回的消息链中不包含参数中给出的消息元素类型
        """
        return self.__class__([i for i in self.content if not isinstance(i, types)])

    def include(self, *types: type[Element]) -> Self:
        """将只在给出的消息元素类型中符合的消息元素重新包装为一个新的消息链
        Args:
            *types (type[Element]): 将只包含在内的消息元素类型
        Returns:
            MessageChain: 返回的消息链中只包含参数中给出的消息元素类型
        """
        return self.__class__([i for i in self.content if isinstance(i, types)])

    def split(self, pattern: str = " ", raw_string: bool = False) -> list[Self]:
        """和 `str.split` 差不多, 提供一个字符串, 然后返回分割结果.

        Args:
            pattern (str): 分隔符. 默认为单个空格.
            raw_string (bool): 是否要包含 "空" 的文本元素.

        Returns:
            list[Self]: 分割结果, 行为和 `str.split` 差不多.
        """

        result: list[Self] = []
        tmp = []
        for element in self.content:
            if isinstance(element, Text):
                split_result = element.text.split(pattern)
                for index, split_str in enumerate(split_result):
                    if tmp and index > 0:
                        result.append(self.__class__(tmp))
                        tmp = []
                    if split_str or raw_string:
                        tmp.append(self._text_class(split_str))
            else:
                tmp.append(element)
        if tmp:
            result.append(self.__class__(tmp))
            tmp = []
        return result

    def __repr__(self) -> str:
        return f"MessageChain({self.content!r})"

    def __iter__(self) -> Iterator[Element]:
        yield from self.content

    def __len__(self) -> int:
        return len(self.content)

    def startswith(self, string: str) -> bool:
        """判断消息链是否以给出的字符串开头

        Args:
            string (str): 字符串

        Returns:
            bool: 是否以给出的字符串开头
        """

        if not self.content or not isinstance(self.content[0], Text):
            return False
        return self.content[0].text.startswith(string)

    def endswith(self, string: str) -> bool:
        """判断消息链是否以给出的字符串结尾

        Args:
            string (str): 字符串

        Returns:
            bool: 是否以给出的字符串结尾
        """

        if not self.content or not isinstance(self.content[-1], Text):
            return False
        return self.content[-1].text.endswith(string)

    def only(self, *element_classes: type[Element]) -> bool:
        """判断消息链中是否只含有特定类型元素.

        Args:
            *element_classes (type[Element]): 元素类型

        Returns:
            bool: 判断结果
        """
        return all(isinstance(i, element_classes) for i in self.content)

    def append(self, element: Element | str, copy: bool = False) -> Self:
        """
        向消息链最后追加单个元素

        Args:
            element (Element): 要添加的元素
            copy (bool): 是否要在副本上修改.

        Returns:
            MessageChain: copy = True 时返回副本, 否则返回自己的引用.
        """
        chain_ref = self.copy() if copy else self
        if isinstance(element, str):
            element = self._text_class(element)
        chain_ref.content.append(element)
        return chain_ref

    def extend(
        self,
        *content: Self | Element | list[Element | str],
        copy: bool = False,
    ) -> Self:
        """
        向消息链最后添加元素/元素列表/消息链

        Args:
            *content (MessageChain | Element | list[Element | str]): 要添加的元素/元素容器.
            copy (bool): 是否要在副本上修改.

        Returns:
            MessageChain: copy = True 时返回副本, 否则返回自己的引用.
        """
        result = []
        for i in content:
            if isinstance(i, Element):
                result.append(i)
            elif isinstance(i, str):
                result.append(self._text_class(i))
            elif isinstance(i, MessageChain):
                result.extend(i.content)
            else:
                for e in i:
                    if isinstance(e, str):
                        result.append(self._text_class(e))
                    else:
                        result.append(e)
        if copy:
            return self.__class__(deepcopy(self.content) + result)
        self.content.extend(result)
        return self

    def empty(self) -> bool:
        """
        判断消息链是否为空，包括判断是否仅包含空字符串。

        Returns:
            bool: 判断结果。
        """

        return not bool(self.content and str(self))

    def copy(self) -> Self:
        """
        拷贝本消息链.

        Returns:
            MessageChain: 拷贝的副本.
        """
        return self.__class__(deepcopy(self.content))

    def index(self, element_type: type[Element]) -> int | None:
        """
        寻找第一个特定类型的元素, 并返回其下标.

        Args:
            element_type (type[Element]): 元素或元素类型

        Returns:
            int | None: 元素下标, 若未找到则为 None.

        """
        return next((i for i, e in enumerate(self.content) if isinstance(e, element_type)), None)

    def count(self, element: type[Element] | Element) -> int:
        """
        统计共有多少个指定的元素.

        Args:
            element (type[Element] | Element): 元素或元素类型

        Returns:
            int: 元素数量
        """
        if isinstance(element, Element):
            return sum(i == element for i in self.content)
        return sum(isinstance(i, element) for i in self.content)

    def index_sub(self, sub: MessageChain | Sequence[str | Element]) -> list[int]:
        """判断消息链是否含有子链. 使用 KMP 算法.

        Args:
            sub (MessageChain | Sequence[str | Element]): 要判断的子链.

        Returns:
            List[int]: 所有找到的下标.
        """

        def unzip(seq: Sequence[str | Element]) -> list[str | Element]:
            res: list[str | Element] = []
            for e in seq:
                if isinstance(e, Text):
                    res.extend(e.text)
                elif isinstance(e, str):
                    res.extend(e)
                else:
                    res.append(e)
            return res

        pattern: list[str | Element] = unzip(sub.content) if isinstance(sub, MessageChain) else unzip(sub)

        match_target: list[str | Element] = unzip(self.content)

        if len(match_target) < len(pattern):
            return []

        fallback: list[int] = [0 for _ in pattern]
        current_fb: int = 0  # current fallback index
        for i in range(1, len(pattern)):
            while current_fb and pattern[i] != pattern[current_fb]:
                current_fb = fallback[current_fb - 1]
            if pattern[i] == pattern[current_fb]:
                current_fb += 1
            fallback[i] = current_fb

        match_index: list[int] = []
        ptr = 0
        for i, e in enumerate(match_target):
            while ptr and e != pattern[ptr]:
                ptr = fallback[ptr - 1]
            if e == pattern[ptr]:
                ptr += 1
            if ptr == len(pattern):
                match_index.append(i - ptr + 1)
                ptr = fallback[ptr - 1]
        return match_index

    def removeprefix(self, prefix: str, *, copy: bool = True) -> Self:
        """移除消息链前缀.

        Args:
            prefix (str): 要移除的前缀.
            copy (bool, optional): 是否在副本上修改, 默认为 True.

        Returns:
            MessageChain: 修改后的消息链, 若未移除则原样返回.
        """
        elements = deepcopy(self.content) if copy else self.content
        if not elements:
            return self.copy() if copy else self
        elem = elements[0]
        if not isinstance(elem, Text):
            return self.copy() if copy else self
        if elem.text.startswith(prefix):
            elem.text = elem.text[len(prefix) :]
            if not elem.text:
                elements.pop(0)
        if copy:
            return self.__class__(elements)
        self.content.clear()
        self.content.extend(elements)
        return self

    def removesuffix(self, suffix: str, *, copy: bool = True) -> Self:
        """移除消息链后缀.

        Args:
            suffix (str): 要移除的后缀.
            copy (bool, optional): 是否在副本上修改, 默认为 True.

        Returns:
            MessageChain: 修改后的消息链, 若未移除则原样返回.
        """
        elements = deepcopy(self.content) if copy else self.content
        if not elements:
            return self.copy() if copy else self
        elem = elements[-1]
        if not isinstance(elem, Text):
            return self.copy() if copy else self
        if elem.text.endswith(suffix):
            elem.text = elem.text[: -len(suffix)]
            if not elem.text:
                elements.pop(-1)
        if copy:
            return self.__class__(elements)
        self.content.clear()
        self.content.extend(elements)
        return self

    def strip(self, *elements: str | type[Element] | Element, copy: bool = True) -> Self:
        return self.lstrip(*elements, copy=copy).rstrip(*elements, copy=copy)

    def lstrip(self, *elements: str | type[Element] | Element, copy: bool = True) -> Self:
        types = [i for i in elements if not isinstance(i, str)] or []
        chars = "".join([i for i in elements if isinstance(i, str)]) or None
        content = deepcopy(self.content) if copy else self.content
        if not content:
            return self.copy() if copy else self
        while content:
            elem = content[0]
            if elem in types or elem.__class__ in types:
                content.pop(0)
            elif isinstance(elem, Text):
                text = elem.text.lstrip(chars)
                if not text:
                    content.pop(0)
                    continue
                elem.text = text
                break
            else:
                break
        if copy:
            return self.__class__(content)
        self.content.clear()
        self.content.extend(content)
        return self

    def rstrip(self, *elements: str | type[Element] | Element, copy: bool = True) -> Self:
        types = [i for i in elements if not isinstance(i, str)] or []
        chars = "".join([i for i in elements if isinstance(i, str)]) or None
        content = deepcopy(self.content) if copy else self.content
        if not content:
            return self.copy() if copy else self
        while content:
            elem = content[-1]
            if elem in types or elem.__class__ in types:
                content.pop(-1)
            elif isinstance(elem, Text):
                text = elem.text.rstrip(chars)
                if not text:
                    content.pop(-1)
                    continue
                elem.text = text
                break
            else:
                break
        if copy:
            return self.__class__(content)
        self.content.clear()
        self.content.extend(content)
        return self

    def replace(
        self,
        old: MessageChain | list[Element],
        new: MessageChain | list[Element],
    ) -> Self:
        """替换消息链中的一部分. (在副本上操作)

        Args:
            old (MessageChain): 要替换的消息链.
            new (MessageChain): 替换后的消息链.

        Returns:
            MessageChain: 修改后的消息链, 若未替换则原样返回.
        """
        if not isinstance(old, MessageChain):
            old = MessageChain(old)
        if not isinstance(new, MessageChain):
            new = MessageChain(new)
        index_list: list[int] = self.index_sub(old)

        def unzip(chain: MessageChain) -> list[str | Element]:
            unzipped: list[str | Element] = []
            for e in chain.content:
                if isinstance(e, Text):
                    unzipped.extend(e.text)
                else:
                    unzipped.append(e)
            return unzipped

        unzipped_new: list[str | Element] = unzip(new)
        unzipped_old: list[str | Element] = unzip(old)
        unzipped_self: list[str | Element] = unzip(self)
        unzipped_result: list[str | Element] = []
        last_end: int = 0
        for start in index_list:
            unzipped_result.extend(unzipped_self[last_end:start])
            last_end = start + len(unzipped_old)
            unzipped_result.extend(unzipped_new)
        unzipped_result.extend(unzipped_self[last_end:])

        # Merge result
        result_list: list[Element] = []
        char_stk: list[str] = []
        for v in unzipped_result:
            if isinstance(v, str):
                char_stk.append(v)
            else:
                result_list.append(self._text_class("".join(char_stk)))
                char_stk = []
                result_list.append(v)
        if char_stk:
            result_list.append(self._text_class("".join(char_stk)))
        return self.__class__(result_list)

    def __add__(self, content: MessageChain | list[Element] | Element | str) -> Self:
        if isinstance(content, str):
            content = self._text_class(content)
        if isinstance(content, Element):
            content = [content]
        if isinstance(content, MessageChain):
            content = content.content
        return type(self)(self.content + content)

    def __radd__(self, content: MessageChain | list[Element] | Element | str) -> Self:
        if isinstance(content, str):
            content = self._text_class(content)
        if isinstance(content, Element):
            content = [content]
        if isinstance(content, MessageChain):
            content = content.content
        return type(self)(content + self.content)

    def __iadd__(self, content: MessageChain | list[Element] | Element | str) -> Self:
        if isinstance(content, str):
            content = self._text_class(content)
        if isinstance(content, Element):
            content = [content]
        if isinstance(content, MessageChain):
            content = content.content
        self.content.extend(content)
        return self

    def __bool__(self):
        return bool(self.content and str(self))

    def __eq__(self, other):
        if not isinstance(other, MessageChain):
            return False
        return other.content == self.content



from __future__ import annotations

import asyncio
import logging
import re
from ssl import VerifyFlags, VerifyMode
from typing import Any, TypedDict

from hypercorn.asyncio import serve
from hypercorn.config import Config
from hypercorn.logging import Logger
from hypercorn.typing import ResponseSummary, WWWScope
from launart import Launart, Service
from launart.status import Phase
from launart.utilles import any_completed
from loguru import logger

from . import asgitypes
from .common import empty_asgi_handler
from .middleware import DispatcherMiddleware


class HypercornOptions(TypedDict, total=False):
    insecure_bind: str | list[str]
    """default: []"""
    quic_bind: str | list[str]
    """default: []"""
    root_path: str
    """default: ''"""

    access_log_format: str
    """default: '%(h)s %(l)s %(u)s %(t)s "%(r)s" %(s)s %(b)s "%(f)s" "%(a)s"'"""
    accesslog: logging.Logger | str | None
    """default: None"""
    alpn_protocols: list[str]
    """default: ['h2', 'http/1.1']"""
    alt_svc_headers: list[str]
    """default: []"""
    backlog: int
    """default: 100"""
    ca_certs: str | None
    """default: None"""
    certfile: str | None
    """default: None"""
    ciphers: str
    """default: 'ECDHE+AESGCM'"""
    dogstatsd_tags: str
    """default: ''"""
    errorlog: logging.Logger | str | None
    """default: '-'"""
    graceful_timeout: float
    """default: 3.0"""
    read_timeout: int | None
    """default: None"""
    group: int | None
    """default: None"""
    h11_max_incomplete_size: int
    """default: 16 * 1024"""
    h11_pass_raw_headers: bool
    """default: False"""
    h2_max_concurrent_streams: int
    """default: 100"""
    h2_max_header_list_size: int
    """default: 2 ** 16"""
    h2_max_inbound_frame_size: int
    """default: 2 ** 14"""
    include_date_header: bool
    """default: True"""
    include_server_header: bool
    """default: True"""
    keep_alive_timeout: float
    """default: 5.0"""
    keep_alive_max_requests: int
    """default: 1000"""
    keyfile: str | None
    """default: None"""
    keyfile_password: str | None
    """default: None"""
    logger_class: type
    """default: hypercorn.logging.Logger"""
    logconfig: str | None
    """default: None"""
    logconfig_dict: dict | None
    """default: None"""
    loglevel: str
    """default: 'INFO'"""
    max_app_queue_size: int
    """default: 10"""
    max_requests: int | None
    """default: None"""
    max_requests_jitter: int
    """default: 0"""
    pid_path: str | None
    """default: None"""
    server_names: list[str]
    """default: []"""
    shutdown_timeout: float
    """default: 60.0"""
    ssl_handshake_timeout: float
    """default: 60.0"""
    startup_timeout: float
    """default: 60.0"""
    statsd_host: str | None
    """default: None"""
    statsd_prefix: str
    """default: ''"""
    umask: int | None
    """default: None"""
    use_reloader: bool
    """default: False"""
    user: int | None
    """default: None"""
    verify_flags: VerifyFlags | None
    """default: None"""
    verify_mode: VerifyMode | None
    """default: None"""
    websocket_max_message_size: int
    """default: 16 * 1024 * 1024"""
    websocket_ping_interval: float | None
    """default: None"""
    worker_class: str
    """default: 'asyncio'"""
    wsgi_max_body_size: int
    """default: 16 * 1024 * 1024"""


class LoguruLogger(Logger):
    def __init__(self, config: Config) -> None:
        super().__init__(config)
        self.access_template = re.sub(r"%\(([^)]+)\)s", r"{\1}", self.access_log_format)
        self.access_template = self.access_template.replace("{t} ", "")

    async def access(self, request: WWWScope, response: ResponseSummary, request_time: float) -> None:
        if self.access_logger is not None:
            logger.info(self.access_template.format(**self.atoms(request, response, request_time)))

    async def critical(self, message: str, *args: Any, **kwargs: Any) -> None:
        logger.critical(message, *args, **kwargs)

    async def error(self, message: str, *args: Any, **kwargs: Any) -> None:
        logger.error(message, *args, **kwargs)

    async def warning(self, message: str, *args: Any, **kwargs: Any) -> None:
        logger.warning(message, *args, **kwargs)

    async def info(self, message: str, *args: Any, **kwargs: Any) -> None:
        logger.info(message, *args, **kwargs)

    async def debug(self, message: str, *args: Any, **kwargs: Any) -> None:
        logger.debug(message, *args, **kwargs)

    async def exception(self, message: str, *args: Any, **kwargs: Any) -> None:
        logger.exception(message, *args, **kwargs)

    async def log(self, level: int, message: str, *args: Any, **kwargs: Any) -> None:
        logger.log(level, message, *args, **kwargs)


class HypercornASGIService(Service):
    id = "asgi.service/hypercorn"

    middleware: DispatcherMiddleware
    host: str
    port: int

    def __init__(
        self,
        host: str,
        port: int,
        mounts: dict[str, asgitypes.ASGI3Application] | None = None,
        options: HypercornOptions | None = None,
        patch_logger: bool = True,
    ):
        self.host = host
        self.port = port
        self.middleware = DispatcherMiddleware(mounts or {"\0\0\0": empty_asgi_handler})
        self.options = options or {}
        if patch_logger:
            self.options["logger_class"] = LoguruLogger  # type: ignore
        super().__init__()

    @property
    def required(self):
        return set()

    @property
    def stages(self) -> set[Phase]:
        return {"preparing", "blocking", "cleanup"}

    async def launch(self, manager: Launart) -> None:
        async with self.stage("preparing"):
            shutdown_trigger = asyncio.Event()
            serve_task = asyncio.create_task(
                serve(
                    self.middleware,  # type: ignore
                    Config.from_mapping(bind=f"{self.host}:{self.port}", **self.options),
                    shutdown_trigger=shutdown_trigger.wait,
                )
            )

        async with self.stage("blocking"):
            await any_completed(serve_task, manager.status.wait_for_sigexit())

        async with self.stage("cleanup"):
            logger.warning("trying to shutdown hypercorn server...")
            shutdown_trigger.set()
            try:
                await asyncio.wait_for(serve_task, timeout=5.0)
            except asyncio.TimeoutError:
                logger.warning("timeout, force exit hypercorn server...")



from collections.abc import Sequence
from typing import Any, ClassVar, Literal, TypeVar, cast

from launart import Launart, Service
from loguru import logger
from sqlalchemy import Table
from sqlalchemy.engine.result import Result
from sqlalchemy.engine.url import URL
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker, create_async_engine
from sqlalchemy.ext.asyncio.engine import AsyncEngine
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.sql.base import Executable
from sqlalchemy.sql.selectable import TypedReturnsRows

from ..utils import get_subclasses
from .model import Base
from .types import EngineOptions

T_Row = TypeVar("T_Row", bound=DeclarativeBase)


class SqlalchemyService(Service):
    id: str = "database/sqlalchemy"
    base_class: ClassVar[type[DeclarativeBase]] = Base
    engines: dict[str, AsyncEngine]
    session_factory: async_sessionmaker[AsyncSession]

    def __init__(
        self,
        url: str | URL,
        engine_options: EngineOptions | None = None,
        session_options: dict[str, Any] | None = None,
        binds: dict[str, str | URL] | None = None,
        create_table_at: Literal["preparing", "prepared", "blocking"] = "preparing",
    ) -> None:
        if engine_options is None:
            engine_options = {"echo": "debug", "pool_pre_ping": True}
        self.engines = {"": create_async_engine(url, **engine_options)}
        for key, bind_url in (binds or {}).items():
            self.engines[key] = create_async_engine(bind_url, **engine_options)
        self.create_table_at = create_table_at
        self.session_options = session_options or {"expire_on_commit": False}
        super().__init__()

    @property
    def required(self) -> set[str]:
        return set()

    @property
    def stages(self) -> set[Literal["preparing", "blocking", "cleanup"]]:
        return {"preparing", "blocking", "cleanup"}

    async def initialize(self):
        _binds = {}
        binds = {}

        for model in set(get_subclasses(self.base_class)):
            table: Table | None = getattr(model, "__table__", None)

            if table is None or (bind_key := table.info.get("bind_key")) is None:
                continue

            if bind_key in self.engines:
                _binds[model] = self.engines[bind_key]
                binds.setdefault(bind_key, []).append(model)
            else:
                _binds[model] = self.engines[""]
                binds.setdefault("", []).append(model)

        self.session_factory = async_sessionmaker(self.engines[""], binds=_binds, **self.session_options)
        return binds

    def get_session(self, **local_kw):
        return self.session_factory(**local_kw)

    async def launch(self, manager: Launart):
        binds: dict[str, list[type[Base]]] = {}

        async with self.stage("preparing"):
            logger.info("Initializing database...")
            if self.create_table_at == "preparing":
                binds = await self.initialize()
                logger.success("Database initialized!")
                for key, models in binds.items():
                    async with self.engines[key].begin() as conn:
                        await conn.run_sync(
                            self.base_class.metadata.create_all, tables=[m.__table__ for m in models], checkfirst=True
                        )
                logger.success("Database tables created!")

        if self.create_table_at != "preparing":
            binds = await self.initialize()
            logger.success("Database initialized!")
        if self.create_table_at == "prepared":
            for key, models in binds.items():
                async with self.engines[key].begin() as conn:
                    await conn.run_sync(
                        self.base_class.metadata.create_all, tables=[m.__table__ for m in models], checkfirst=True
                    )
            logger.success("Database tables created!")

        async with self.stage("blocking"):
            if self.create_table_at == "blocking":
                for key, models in binds.items():
                    async with self.engines[key].begin() as conn:
                        await conn.run_sync(
                            self.base_class.metadata.create_all, tables=[m.__table__ for m in models], checkfirst=True
                        )
                logger.success("Database tables created!")
            await manager.status.wait_for_sigexit()
        async with self.stage("cleanup"):
            for engine in self.engines.values():
                await engine.dispose(close=True)

    async def execute(self, sql: Executable) -> Result:
        """执行 SQL 命令"""
        async with self.get_session() as session:
            return await session.execute(sql)

    async def select_all(self, sql: TypedReturnsRows[tuple[T_Row]]) -> Sequence[T_Row]:
        async with self.get_session() as session:
            result = await session.scalars(sql)
        return result.all()

    async def select_first(self, sql: TypedReturnsRows[tuple[T_Row]]) -> T_Row | None:
        async with self.get_session() as session:
            result = await session.scalars(sql)
        return cast("T_Row | None", result.first())

    async def add(self, row: Base) -> None:
        async with self.get_session() as session:
            session.add(row)
            await session.commit()
            await session.refresh(row)

    async def add_many(self, rows: Sequence[Base]):
        async with self.get_session() as session:
            session.add_all(rows)
            await session.commit()
            for row in rows:
                await session.refresh(row)

    async def update_or_add(self, row: Base):
        async with self.get_session() as session:
            await session.merge(row)
            await session.commit()
            await session.refresh(row)

    async def delete_exist(self, row: Base):
        async with self.get_session() as session:
            await session.delete(row)

    async def delete_many_exist(self, rows: Sequence[Base]):
        async with self.get_session() as session:
            for row in rows:
                await session.delete(row)



from graia.amnesia.message import MessageChain
from graia.amnesia.message.element import Text, Unknown


def test_init():
    assert MessageChain(["123", "456"]).merge(copy=False) == MessageChain("123456")


def test_bool():
    assert not MessageChain([])
    assert not MessageChain("")


def test_strip():
    msg = MessageChain([Text("    123"), Unknown("at", {"id": 1})])

    assert msg.lstrip() == MessageChain([Text("123"), Unknown("at", {"id": 1})])
    assert msg.rstrip(Unknown) == MessageChain("    123")
    assert msg.strip(" ", Unknown) == MessageChain("123")
    assert msg.strip(Text, Unknown).empty()


def test_replace():
    msg = MessageChain([Text("    123"), Unknown("at", {"id": 1})])

    assert msg.replace([Text("123")], [Text("456")]) == MessageChain([Text("    456"), Unknown("at", {"id": 1})])


def test_remove():
    msg = MessageChain([Text("123"), Unknown("at", {"id": 1}), Text("456")])

    assert msg.removeprefix("123") == MessageChain([Unknown("at", {"id": 1}), Text("456")])
    assert msg.removesuffix("456") == MessageChain([Text("123"), Unknown("at", {"id": 1})])



from __future__ import annotations

import random
import string
from collections.abc import Callable, Hashable
from typing import TypeVar, Union

T = TypeVar("T")
H = TypeVar("H", bound=Hashable)

PriorityType = Union[
    "set[T]",
    "dict[T, int | float]",
    "tuple[set[T] | dict[T, int | float], ...]",
]


def priority_strategy(
    items: list[T],
    getter: Callable[
        [T],
        PriorityType[H],
    ],
) -> dict[H, T]:
    result = {}
    _cache = {}

    def _raise_conflict(content):
        raise ValueError(
            f"{content} which is an unlocated item is already existed, and it conflicts with {result[content]}"
        )

    def _raise_existed(content):
        raise ValueError(f"{content} is already existed, and it conflicts with {result[content]}, an unlocated item.")

    def _handle(pattern):
        if isinstance(pattern, set):
            for content in pattern:
                if content in _cache:
                    _raise_conflict(content)
                _cache[content] = ...
                result[content] = item
        elif isinstance(pattern, dict):
            for content, priority in pattern.items():
                if content in _cache:
                    if _cache[content] is ...:
                        _raise_existed(content)
                    if priority is ...:
                        _raise_conflict(content)
                    if _cache[content] < priority:
                        _cache[content] = priority
                        result[content] = item
                else:
                    _cache[content] = priority
                    result[content] = item
        else:
            raise TypeError(f"{pattern} is not a valid pattern.")

    for item in items:
        pattern = getter(item)
        if isinstance(pattern, (dict, set)):
            _handle(pattern)
        elif isinstance(pattern, tuple):
            for subpattern in pattern:
                _handle(subpattern)
        else:
            raise TypeError(f"{pattern} is not a valid pattern.")
    return result


T = TypeVar("T")


class Registrar(dict):
    def register(self, key):
        def decorator(method):
            self[key] = method
            return method

        return decorator

    def decorate(self, attr):
        def decorator(cls: type[T]) -> type[T]:
            getattr(cls, attr).update(self)
            return cls

        return decorator


def random_id(length=12):
    return "".join(random.choices(string.ascii_letters + string.digits, k=length))



import asyncio
import logging
import os
from collections.abc import Awaitable, Callable
from typing import IO, Any, Literal, TypedDict

from launart import Launart, Service
from launart.status import Phase
from launart.utilles import any_completed
from loguru import logger
from uvicorn import Config, Server
from uvicorn.config import LOG_LEVELS, HTTPProtocolType, LifespanType, WSProtocolType

from ..utils import LoguruHandler
from . import asgitypes
from .common import empty_asgi_handler
from .middleware import DispatcherMiddleware


class WithoutSigHandlerServer(Server):
    def install_signal_handlers(self) -> None:
        pass


class UvicornOptions(TypedDict, total=False):
    uds: str | None
    """default: None"""
    fd: int | None
    """default: None"""
    loop: Literal["none", "auto", "asyncio", "uvloop"]
    """default: 'auto'"""
    http: type[asyncio.Protocol] | HTTPProtocolType
    """default: 'auto'"""
    ws: type[asyncio.Protocol] | WSProtocolType
    """default: 'auto'"""
    ws_max_size: int
    """default: 16 * 1024 * 1024"""
    ws_max_queue: int
    """default: 32"""
    ws_ping_interval: float | None
    """default: 20.0"""
    ws_ping_timeout: float | None
    """default: 20.0"""
    ws_per_message_deflate: bool
    """default: True"""
    lifespan: LifespanType
    """default: 'auto'"""
    env_file: str | os.PathLike[str] | None
    """default: None"""
    log_config: dict[str, Any] | str | IO[Any] | None
    """default: LOGGING_CONFIG"""
    log_level: str | int | None
    """default: None"""
    access_log: bool
    """default: True"""
    use_colors: bool | None
    """default: None"""
    # interface: InterfaceType
    # """default: 'auto'"""
    reload: bool
    """default: False"""
    reload_dirs: list[str] | str | None
    """default: None"""
    reload_delay: float
    """default: 0.25"""
    reload_includes: list[str] | str | None
    """default: None"""
    reload_excludes: list[str] | str | None
    """default: None"""
    workers: int | None
    """default: None"""
    proxy_headers: bool
    """default: True"""
    server_header: bool
    """default: True"""
    date_header: bool
    """default: True"""
    forwarded_allow_ips: list[str] | str | None
    """default: None"""
    root_path: str
    """default: ''"""
    limit_concurrency: int | None
    """default: None"""
    limit_max_requests: int | None
    """default: None"""
    backlog: int
    """default: 2048"""
    timeout_keep_alive: int
    """default: 5"""
    timeout_notify: int
    """default: 30"""
    timeout_graceful_shutdown: int | None
    """default: None"""
    callback_notify: Callable[..., Awaitable[None]] | None
    """default: None"""
    ssl_keyfile: str | os.PathLike[str] | None
    """default: None"""
    ssl_certfile: str | os.PathLike[str] | None
    """default: None"""
    ssl_keyfile_password: str | None
    """default: None"""
    ssl_version: int
    """default: ssl.PROTOCOL_TLS"""
    ssl_cert_reqs: int
    """default: ssl.CERT_NONE"""
    ssl_ca_certs: str | None
    """default: None"""
    ssl_ciphers: str
    """default: 'TLSv1'"""
    headers: list[tuple[str, str]] | None
    """default: None"""
    h11_max_incomplete_event_size: int | None
    """default: None"""


class UvicornASGIService(Service):
    id = "asgi.service/uvicorn"

    middleware: DispatcherMiddleware
    host: str
    port: int

    def __init__(
        self,
        host: str,
        port: int,
        mounts: dict[str, asgitypes.ASGI3Application] | None = None,
        options: UvicornOptions | None = None,
        patch_logger: bool = True,
    ):
        self.host = host
        self.port = port
        self.patch_logger = patch_logger
        self.middleware = DispatcherMiddleware(mounts or {"\0\0\0": empty_asgi_handler})
        self.options: UvicornOptions = options or {}
        super().__init__()

    @property
    def required(self):
        return set()

    @property
    def stages(self) -> set[Phase]:
        return {"preparing", "blocking", "cleanup"}

    async def launch(self, manager: Launart) -> None:
        async with self.stage("preparing"):
            self.server = WithoutSigHandlerServer(
                Config(self.middleware, host=self.host, port=self.port, factory=False, **self.options)
            )
            if self.patch_logger:
                self._patch_logger()
            serve_task = asyncio.create_task(self.server.serve())

        async with self.stage("blocking"):
            await any_completed(serve_task, manager.status.wait_for_sigexit())

        async with self.stage("cleanup"):
            logger.warning("try to shutdown uvicorn server...")
            self.server.should_exit = True
            await any_completed(serve_task, asyncio.sleep(5))
            if not serve_task.done():
                logger.warning("timeout, force exit uvicorn server...")

    def _patch_logger(self) -> None:
        log_level = 20
        if "log_level" in self.options and (_log_level := self.options["log_level"]) is not None:
            if isinstance(_log_level, str):
                log_level = LOG_LEVELS[_log_level]
            else:
                log_level = _log_level
        PATCHES = ["uvicorn.error", "uvicorn.asgi", "uvicorn"]
        if "access_log" not in self.options or self.options.get("access_log", True):
            PATCHES.append("uvicorn.access")
        for name in PATCHES:
            target = logging.getLogger(name)
            target.handlers = [LoguruHandler()]
            target.propagate = False
            target.setLevel(log_level)
